@@ -9,7 +9,7 @@ use List::MoreUtils qw(before);
 my $root = '/example-election';
 my $zk   = ZooKeeper->new(hosts => 'localhost:2181');
 
-$zk->create($root, persistent => 1) unless $zk->exists($root);
+$zk->create($root) unless $zk->exists($root);
 join_group();
 
 # make sure SIGINT cleanly destroys zookeeper connection
@@ -59,7 +59,8 @@ sub elect_leader {
 }
 
 sub join_group {
-    my $me = node_from_path($zk->create("$root/n-", sequential => 1));
+    my $path = $zk->create("$root/n-", ephemeral => 1, sequential => 1);
+    my $me   = node_from_path($path);
     elect_leader($me);
 }
 
